"""Copyright (c) 2023 Scott Staniewicz. All rights reserved.

geepers: Download GPS data and compare to InSAR
"""

# flake8: noqa
# _version.py is autogenerated by setuptools_scm
from __future__ import annotations

from geepers._version import version as __version__
from .gps import *
from .io import *

__all__ = ["__version__"]
